import socket, time, hashlib, os
from threading import Thread

# for binary files use mode='rb' or mode='wb'
def send_file(sock, file, mode="rb", bufsize=4096):
    fsize = int(os.path.getsize(file))
    hexsize = hex(fsize)[2:]  # convert to hex and chop the leading '0x'
    header = (8 - len(hexsize)) * '0' + hexsize
    sock.sendall(header)
    f = open(file, mode)  # open file for reading in binary mode
    while True:       
        data = f.read(bufsize)
        if not data:
            f.close()
            break
        rcode = sock.sendall(data)
        if not rcode is None:
            print "Exceptional socket sendall return code:", rcode
            return rcode

# Dump socket output (sock.recv) to a local file
# Note the return value is two integers: the size that was
# sent and the real file size. Clients can check if these differ and
# issue a warning and act in any required way
def recv_to_file(sock, filename, mode='wb', bufsize=4096):
    hexstr = "0x" + recv_fixed_size(sock,8)
    file_size = int(hexstr, 16)
    f = open(filename, mode)
    s = 0
    while s<file_size:
        bufsz = min(bufsize, file_size-s)
        data = sock.recv(bufsz)
        s += len(data)
        if not data:
            print "Failed to get all file data !? Socket closed prematurely !?"
            break
        f.write(data)
    f.close()
    return s, file_size

def safe_send(sock, message):
    i = 0
    n = len(message)
    while i < n:
        sent = sock.send(message[i:])
        if sent == 0:
            raise RuntimeError("socket connection broken")
        i += sent

# This is good only if the message expected size is known ahead
# otherwise you will get a runtime error even if you received the full message!
def recv_fixed_size(sock, expected_size, bufsize=0):
    if bufsize == 0:
        bufsize = min(expected_size, 4096)
    message = ""
    while len(message) < expected_size:
        chunk = sock.recv(bufsize)
        if chunk == "":
            raise RuntimeError("socket connection broken")
        message += chunk
    return message

# IDEA: we want to send a message with its size decalred ahead, so the receiver will know
# in advance how much data is going to arrive.
# How do we do that?
# Assuming that our message is not too big its size can be packed in an 8 bytes hex header !
# For example, if message is: "Hello Web Wide World"
# then our size headr = "0000014"
# (length of "Hello Web Wide World" is 20 which in hex is: 0x14)
# So we will send this message: "0000014Hello Web Wide World"
# The receiver on the other side of the channel, knowing the format that we send him (by agrreement),
# will extract the first 8 bytes from our message and will know how much left to get.
# This is the send_size function that goes with recv_size
def send_size(sock, message):
    size_string = hex(len(message))[2:]  # convert message length to hex and chop the leading '0x'
    data = (8 - len(size_string)) * '0' + size_string + message
    sock.sendall(data)

# The receiver gets the first 8 bytes, adds a "0x" prefix, and converts the hex to decimal
def recv_size(sock, bufsize=0):
    hexstr = "0x" + recv_fixed_size(sock,8)
    size = int(hexstr, 16)
    return recv_fixed_size(sock, size, bufsize)

# This is good only if the message expected size is known ahead!
# otherwise you may block or get a runtime error even if you received the full message!
def recv_fixed_size_to_file(sock, size, file, mode="wb", bufsize=0):
    if bufsize == 0:
        bufsize = min(size, 4096)
    f = open(file, mode)
    curr_size = 0
    while curr_size < size:
        data = sock.recv(bufsize)
        if data == "":
            #raise RuntimeError("socket connection broken")
            break
        f.write(data)
        curr_size += len(data)
    f.close()
    return curr_size

def recv_delimited_message(sock, limit='\n'):
    message = ""
    while True:
        char = sock.recv(1)
        if char == "":
            return None
        if char == limit:
            break
        else:
            message += char
    return message

# Thread for sending a file through a service socket which is
# closed immediately after transferring the file
class SendFileThread(Thread):
    def __init__(self, sock, file, mode="rb", bufsize=4096):
        Thread.__init__(self)
        self.sock = sock
        self.file = file
        self.mode = mode
        self.bufsize = bufsize

    def run(self):
        start = time.time()%1000
        print "Started Thread %s: File = %s, Time = %.2f" % (self.getName(), self.file, start)
        fsize = os.path.getsize(self.file)
        self.sock.sendall(str(fsize) + '\n')   # Send file size to client
        ok = recv_fixed_size(self.sock, 2)     # receive file transfer confirmation from client: "OK"
        #print ok                              # more checks are required here if ok != "OK" ! ...
        status = send_file(self.sock, self.file, self.mode, self.bufsize)
        if status is None:
            print "send_file failed: probably cleint socket closed"
        else:
            response = recv_fixed_size(self.sock, 4)  # response is either "DONE" or "FAIL" (4 chars)
            if response == "DONE":
                print "Transfer Confirmed"
            elif response == "FAIL":
                print "Transfer Failed"
        end = time.time()%1000
        print "Completed Thread %s: FILE = %s, TIME = %.2f--%.2f" % (self.getName(), self.file, start, end)
        self.sock.close()

# Thread for receiving a fixed size file through socket
# Warning 1: socket must be closed immediately after getting the file (why?)
# Warning 2: This is just a quick and dirty idea of a class. This is not a final
# and very useful design (which you may need for the BFTP project).
# Feel free to copy and improve for better versions
class RecvFileThread(Thread):
    def __init__(self, sock, size, file, mode="wb", bufsize=4096):
        Thread.__init__(self)
        self.sock = sock
        self.size = size
        self.file = file
        self.mode = mode
        self.bufsize = bufsize

    def run(self):
        start = time.time()%1000
        print "Started Thread %s: File = %s, Time = %.2f" % (self.getName(), self.file, start)
        recv_fixed_size_to_file(self.sock, self.size, self.file, self.mode, self.bufsize)
        #? self.sock.sendall("OK") # depends on protocol design ...
        end = time.time()%1000
        print "Completed Thread %s: FILE = %s, TIME = %.2f--%.2f" % (self.getName(), self.file, start, end)
        if os.path.getsize(self.file) == self.size:
            print "Received file:", self.file
            self.sock.sendall('DONE')
        else:
            print "Receive error: size does not match", self.file
            self.sock.sendall('FAIL')
        self.sock.close()

# return True if files are equal, False if not equal
# using checksum technique, so this is not fully 100% accurate,
# but 100% good enough for small tests
def files_equal(file1, file2):
    d1 = file_digest(file1)
    d2 = file_digest(file2)
    return d1 == d2

# This function takes a large file and returns a short string which represents
# the file "digest" (sometimes also called "signature" or "checksum" of the file)
# With very high probability, if two files have the same digest, then they are identical
def file_digest(file):
    md = hashlib.sha256()
    f = open(file, "r")
    bufsize = 2**24
    while True:
        data = f.read(bufsize)
        if len(data) == 0:
            break
        md.update(data)
    f.close()
    digest = md.hexdigest()
    #size = os.path.getsize(file)
    #digest = digest + ":" + str(size)
    return digest

def getIPaddress():
    sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
    sock.connect(('google.com', 80))
    ip, port = sock.getsockname()
    sock.close()
    return ip




